/*
 *  linux/drivers/fusion/mpiscsi.c
 *      High performance SCSI / Fibre Channel device driver.
 *      For use with PCI adapter(s):
 *          LSI SYM409xx (Fibre Channel)
 *      running LSI Logic Fusion MPT (Message Passing Technology) firmware.
 *
 *  Credits:
 *      This driver would not exist if not for Alan Cox's development
 *      of the linux i2o driver.
 *      A huge debt of gratitude is owed to David S. Miller (DaveM)
 *      for fixing much of the stupid and broken stuff in the early
 *      driver while porting to sparc64 platform.  THANK YOU!
 *
 *  Copyright (c) 1999-2000 LSI Logic Corporation
 *  Originally By: Steven J. Ralston
 *  (mailto:Steve.Ralston@lsil.com)
 *
 *  $Id: mpiscsi.c,v 1.53 2000/09/27 10:17:03 sralston Exp $
 */
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    NO WARRANTY
    THE PROGRAM IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OR
    CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED INCLUDING, WITHOUT
    LIMITATION, ANY WARRANTIES OR CONDITIONS OF TITLE, NON-INFRINGEMENT,
    MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. Each Recipient is
    solely responsible for determining the appropriateness of using and
    distributing the Program and assumes all risks associated with its
    exercise of rights under this Agreement, including but not limited to
    the risks and costs of program errors, damage to or loss of data,
    programs or equipment, and unavailability or interruption of operations.

    DISCLAIMER OF LIABILITY
    NEITHER RECIPIENT NOR ANY CONTRIBUTORS SHALL HAVE ANY LIABILITY FOR ANY
    DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING WITHOUT LIMITATION LOST PROFITS), HOWEVER CAUSED AND
    ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
    TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
    USE OR DISTRIBUTION OF THE PROGRAM OR THE EXERCISE OF ANY RIGHTS GRANTED
    HEREUNDER, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#include <linux/config.h>
#include <linux/version.h>

#if defined(MODULE) && !defined(__GENKSYMS__)
#include <linux/module.h>
#endif

#include <linux/kernel.h>
#include <linux/errno.h>

#include <linux/kdev_t.h>
#include <linux/blkdev.h>
#include <linux/blk.h>		/* for io_request_lock (spinlock) decl */
#include "../scsi/scsi.h"
#include "../scsi/hosts.h"
#include "../scsi/sd.h"

#include "mpi.h"
#include "mpiscsi.h"

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
#define my_NAME     "Fusion MPT SCSI Initiator"
#define my_VERSION  LINUX_MPI_VERSION_COMMON

#ifdef MODULE
MODULE_AUTHOR(MoDuLeAuThOr);
MODULE_DESCRIPTION(my_NAME);
#endif

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*
 *  SCSI Host / Initiator decls...
 */
STATIC        int   mpi_scsi_detect(Scsi_Host_Template *);
STATIC        int   mpi_scsi_release(struct Scsi_Host *host);
STATIC const  char *mpi_scsi_info(struct Scsi_Host *);
/*STATIC        int   mpi_scsi_command(Scsi_Cmnd *);*/
STATIC        int   mpi_scsi_queuecommand(Scsi_Cmnd *, void (*done)(Scsi_Cmnd *));
#ifdef MPI_SCSI_USE_NEW_EH
STATIC        int   mpi_scsi_abort(Scsi_Cmnd *);
STATIC        int   mpi_scsi_bus_reset(Scsi_Cmnd *);
STATIC        int   mpi_scsi_dev_reset(Scsi_Cmnd *);
/*STATIC        int   mpi_scsi_host_reset(Scsi_Cmnd *);*/
#else
STATIC        int   mpi_scsi_old_abort(Scsi_Cmnd *);
STATIC        int   mpi_scsi_old_reset(Scsi_Cmnd *, unsigned int);
#endif
STATIC        void  mpi_scsi_taskmgmt_bh(void *);
STATIC        int   mpi_scsi_bios_param(Disk *, kdev_t, int *);


/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

typedef struct _BIG_SENSE_BUF {
	U8		data[256];
} BIG_SENSE_BUF;

typedef struct _MPI_SCSI_HOST {
	MPI_ADAPTER		 *ioc;
	int			  port;
	struct scsi_cmnd	**ScsiLookup;
	U8			 *SgHunks;
	dma_addr_t		  SgHunksDMA;
	FCDEV_TRACKER		  TargetsQ;
	U32			  qtag_tick;
} MPI_SCSI_HOST;

typedef struct _MPI_SCSI_DEV {
	struct _MPI_SCSI_DEV	 *forw;
	struct _MPI_SCSI_DEV	 *back;
	MPI_ADAPTER		 *ioc;
	int			  sense_sz;
	BIG_SENSE_BUF		  CachedSense;
	unsigned long		  io_cnt;
	unsigned long		  read_cnt;
} MPI_SCSI_DEV;

/*
 *  Other private/forward protos...
 */

static int   ScsiDone(MPI_ADAPTER *ioc, MPI_FRAME_HDR *mf, MPI_FRAME_HDR *r);
static void  report_queue_full(Scsi_Cmnd *sc, SCSIIOReply_t *pScsiReply, SCSIIORequest_t *pScsiReq);
static int   ScsiTaskMgmtComplete(MPI_ADAPTER *ioc, MPI_FRAME_HDR *mf, MPI_FRAME_HDR *r);
static int   scsi_dir(Scsi_Cmnd *cmd);
static void  copy_sense_data(Scsi_Cmnd *sc, MPI_SCSI_HOST *hd, MPI_FRAME_HDR *mf, SCSIIOReply_t *pScsiReply);
static U32   SCPNT_TO_MSGCTX(Scsi_Cmnd *sc);
static int   ScsiErrorReport(IO_Info_t *ioop);


static int       mpi_scsi_hosts = 0;
static atomic_t  queue_depth;

static int       ScsiDoneCtx = -1;
static int       ScsiTaskCtx = -1;

#ifdef MODULE
Scsi_Host_Template mpi_scsi_host_template = MPISCSI;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,28)
static struct proc_dir_entry proc_mpiscsi =
{
	PROC_SCSI_MPI,
	7,
	"mpiscsi",
	S_IFDIR | S_IRUGO | S_IXUGO,
	2
};
#endif

#define SNS_LEN(scp)  sizeof((scp)->sense_buffer)

/*
 *  Stuff to handle single-threading SCSI TaskMgmt
 *  (abort/reset) requests...
 */
static spinlock_t mpiscsi_taskQ_lock = SPIN_LOCK_UNLOCKED;
STATIC MPI_Q_TRACKER mpiscsi_taskQ = {
	(MPI_FRAME_HDR*) &mpiscsi_taskQ,
	(MPI_FRAME_HDR*) &mpiscsi_taskQ
};
STATIC int mpiscsi_taskQ_cnt = 0;
STATIC int mpiscsi_taskQ_bh_active = 0;
STATIC MPI_FRAME_HDR *mpiscsi_active_taskmgmt_mf = NULL;

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	ScsiDone - Main SCSI IO callback routine registered to mpi (base)
 *			driver
 *	@ioc: Pointer to MPI_ADAPTER structure
 *	@mf: Pointer to original MPI request frame
 *	@r: Pointer to MPI reply frame (NULL if TurboReply)
 *
 *	This routine is called from mpi.c::mpi_interrupt() at the completion
 *	of any SCSI IO request.
 *	This routine is registered with the mpi (base) driver at driver
 *	load/init time via the mpi_register() API call.
 *
 *	Returns 1 indicating alloc'd request frame ptr should be freed.
 */
static int
ScsiDone(MPI_ADAPTER *ioc, MPI_FRAME_HDR *mf, MPI_FRAME_HDR *r)
{
	Scsi_Cmnd       *sc;
	MPI_SCSI_HOST   *hd;
	MPI_SCSI_DEV	*mpi_sdev = NULL;
	U16		 req_idx;

	if ((mf == NULL) ||
	    (mf >= MPI_INDEX_2_MFPTR(ioc, ioc->req_depth))) {
		printk(KERN_ERR "mpiscsi: ERROR! NULL or BAD req frame ptr (=%p)!\n", mf);
		return 1;
	}

	hd = (MPI_SCSI_HOST *) ioc->sh->hostdata;
	req_idx = le16_to_cpu(mf->u.frame.hwhdr.msgctxu.fld.req_idx);
	sc = hd->ScsiLookup[req_idx];
	hd->ScsiLookup[req_idx] = NULL;

	dirqprintk((KERN_INFO "mpiscsi: ScsiDone (req:sc:reply=%p:%p:%p)\n", mf, sc, r));

	atomic_dec(&queue_depth);

	/*
	 *  Check for {1st} {IO} completion to "new" device.
	 *  How do we know it's a new device?
	 *  If we haven't set SDpnt->hostdata I guess...
	 */
	if (sc && sc->device) {
		mpi_sdev = (MPI_SCSI_DEV*)sc->device->hostdata;
		if (!mpi_sdev) {
			dprintk((KERN_INFO "mpiscsi: *NEW* SCSI device (%d:%d:%d)!\n",
					   sc->device->id, sc->device->lun, sc->device->channel));
			if ((sc->device->hostdata = kmalloc(sizeof(MPI_SCSI_DEV), GFP_ATOMIC)) == NULL) {
				printk(KERN_ERR "mpiscsi: ERROR: kmalloc(%d) FAILED!\n", (int)sizeof(MPI_SCSI_DEV));
			} else {
				memset(sc->device->hostdata, 0, sizeof(MPI_SCSI_DEV));
				mpi_sdev = (MPI_SCSI_DEV *) sc->device->hostdata;
				mpi_sdev->ioc = ioc;
			}
		} else {
			if (++mpi_sdev->io_cnt && scsi_dir(sc) < 0) {
				if (++mpi_sdev->read_cnt == 3) {
					dprintk((KERN_INFO "mpiscsi: 3rd DATA_IN, CDB[0]=%02x\n",
							   sc->cmnd[0]));
				}
			}
#if 0
			if (mpi_sdev->sense_sz) {
				/*
				 *  Completion of first IO down this path
				 *  *should* invalidate device SenseData...
				 */
				mpi_sdev->sense_sz = 0;
			}
#endif
		}
	}

#if 0
{
	MPI_FRAME_HDR	*mf_chk;

	/* This, I imagine, is a costly check, but...
	 *  If abort/reset active, check to see if this is a IO
	 *  that completed while ABORT/RESET for it is waiting
	 *  on our taskQ!
	 */
	if (! Q_IS_EMPTY(&mpiscsi_taskQ)) {
		// If ABORT for this IO is queued, zap it!
		mf_chk = search_taskQ(1,sc,MPI_SCSITASKMGMT_TASKTYPE_ABORT_TASK);
		if (mf_chk != NULL) {
			sc->result = DID_ABORT << 16;
			spin_lock_irqsave(&io_request_lock, flags);
			sc->scsi_done(sc);
			spin_unlock_irqrestore(&io_request_lock, flags);
			return 1;
		}
	}
}
#endif

	if (r != NULL && sc != NULL) {
		SCSIIOReply_t   *pScsiReply;
		SCSIIORequest_t *pScsiReq;
		U16		 status;

		pScsiReply = (SCSIIOReply_t *) r;
		pScsiReq = (SCSIIORequest_t *) mf;

		status = le16_to_cpu(pScsiReply->IOCStatus) & MPI_IOCSTATUS_MASK;

		dprintk((KERN_NOTICE "mpiscsi: Uh-Oh!  (req:sc:reply=%p:%p:%p)\n", mf, sc, r));
		dprintk((KERN_NOTICE "  IOCStatus=%04xh, SCSIState=%02xh"
				     ", SCSIStatus=%02xh, IOCLogInfo=%08xh\n",
				     status, pScsiReply->SCSIState, pScsiReply->SCSIStatus,
				     le32_to_cpu(pScsiReply->IOCLogInfo)));

		/*
		 *  Look for + dump FCP ResponseInfo[]!
		 */
		if (pScsiReply->SCSIState & MPI_SCSI_STATE_RESPONSE_INFO_VALID) {
			dprintk((KERN_NOTICE "  FCP_ResponseInfo=%08xh\n",
					     le32_to_cpu(pScsiReply->ResponseInfo)));
		}

		switch(status) {
		case MPI_IOCSTATUS_BUSY:			/* 0x0002 */
			/*sc->result = DID_BUS_BUSY << 16;*/		/* YIKES! - Seems to
									 * kill linux interrupt
									 * handler
									 */
			sc->result = STS_BUSY;				/* Try SCSI BUSY! */
			break;

		case MPI_IOCSTATUS_SCSI_RECOVERED_ERROR:	/* 0x0040 */
			/*  Not real sure here...  */
			sc->result = DID_OK << 16;
			break;

		case MPI_IOCSTATUS_SCSI_INVALID_BUS:		/* 0x0041 */
		case MPI_IOCSTATUS_SCSI_INVALID_TARGETID:	/* 0x0042 */
			sc->result = DID_BAD_TARGET << 16;
			break;

		case MPI_IOCSTATUS_SCSI_DEVICE_NOT_THERE:	/* 0x0043 */
			/*  Spoof to SCSI Selection Timeout!  */
			sc->result = DID_NO_CONNECT << 16;
			break;

		case MPI_IOCSTATUS_SCSI_DATA_UNDERRUN:		/* 0x0045 */
			/*
			 *  YIKES!  I just discovered that SCSI IO which
			 *  returns check condition, SenseKey=05 (ILLEGAL REQUEST)
			 *  and ASC/ASCQ=94/01 (LSI Logic RAID vendor specific),
			 *  comes down this path!
			 *  Do upfront check for valid SenseData and give it
			 *  precedence!
			 */
			if (pScsiReply->SCSIState & MPI_SCSI_STATE_AUTOSENSE_VALID) {
				copy_sense_data(sc, hd, mf, pScsiReply);
				sc->result = pScsiReply->SCSIStatus;
				break;
			}

			dprintk((KERN_NOTICE "mpiscsi: sc->underflow={report ERR if < %02xh bytes xfer'd}\n", sc->underflow));
			dprintk((KERN_NOTICE "mpiscsi: ActBytesXferd=%02xh\n", le32_to_cpu(pScsiReply->TransferCount)));

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
			sc->resid = sc->request_bufflen - le32_to_cpu(pScsiReply->TransferCount);
			dprintk((KERN_NOTICE "mpiscsi: SET sc->resid=%02xh\n", sc->resid));
#endif

#if 0
			if (sc->underflow && (le32_to_cpu(pScsiReply->TransferCount) < sc->underflow)) {
				sc->result = DID_ERROR << 16;
				sc->resid = sc->request_bufflen - le32_to_cpu(pScsiReply->TransferCount);
			} else {
				sc->result = 0;
			}
#endif

			/* workaround attempts... */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
			if (sc->resid >= 0x200) {
				// GRRRRR...
				//sc->result = DID_SOFT_ERROR << 16;
				// Try spoofing to BUSY
				sc->result = STS_BUSY;
			} else {
				sc->result = 0;
			}
#else
			sc->result = 0;
#endif
			break;

		case MPI_IOCSTATUS_SCSI_TASK_TERMINATED:	/* 0x0048 */
			sc->result = DID_ABORT << 16;
			break;

		case MPI_IOCSTATUS_SCSI_IOC_TERMINATED:		/* 0x004B */
		case MPI_IOCSTATUS_SCSI_EXT_TERMINATED:		/* 0x004C */
			sc->result = DID_RESET << 16;
			break;

		case MPI_IOCSTATUS_SUCCESS:			/* 0x0000 */
			sc->result = pScsiReply->SCSIStatus;

			if (pScsiReply->SCSIState & MPI_SCSI_STATE_AUTOSENSE_VALID) {
				copy_sense_data(sc, hd, mf, pScsiReply);

				/*  If running agains circa 200003dd 909 MPI f/w,
				 *  may get this (AUTOSENSE_VALID) for actual TASK_SET_FULL
				 *  (QUEUE_FULL) returned from device!  --> get 0x0000?128
				 *  and with SenseBytes set to 0.
				 */
				if (pScsiReply->SCSIStatus == MPI_SCSI_STATUS_TASK_SET_FULL)
					report_queue_full(sc, pScsiReply, pScsiReq);
			}
			else if (pScsiReply->SCSIState & (MPI_SCSI_STATE_AUTOSENSE_FAILED | MPI_SCSI_STATE_NO_SCSI_STATUS)) {
				/*
				 *  What to do?
				 */
				sc->result = DID_SOFT_ERROR << 16;
			}
			else if (pScsiReply->SCSIState & MPI_SCSI_STATE_TERMINATED) {
				/*  Not real sure here either...  */
				sc->result = DID_ABORT << 16;
			}

			if (sc->result == MPI_SCSI_STATUS_TASK_SET_FULL)
				report_queue_full(sc, pScsiReply, pScsiReq);

			break;

		case MPI_IOCSTATUS_INVALID_FUNCTION:		/* 0x0001 */
		case MPI_IOCSTATUS_INVALID_SGL:			/* 0x0003 */
		case MPI_IOCSTATUS_INTERNAL_ERROR:		/* 0x0004 */
		case MPI_IOCSTATUS_RESERVED:			/* 0x0005 */
		case MPI_IOCSTATUS_INSUFFICIENT_RESOURCES:	/* 0x0006 */
		case MPI_IOCSTATUS_INVALID_FIELD:		/* 0x0007 */
		case MPI_IOCSTATUS_INVALID_STATE:		/* 0x0008 */
		case MPI_IOCSTATUS_SCSI_DATA_OVERRUN:		/* 0x0044 */
		case MPI_IOCSTATUS_SCSI_IO_DATA_ERROR:		/* 0x0046 */
		case MPI_IOCSTATUS_SCSI_PROTOCOL_ERROR:		/* 0x0047 */
		case MPI_IOCSTATUS_SCSI_RESIDUAL_MISMATCH:	/* 0x0049 */
		case MPI_IOCSTATUS_SCSI_TASK_MGMT_FAILED:	/* 0x004A */
		default:
			/*
			 *  What to do?
			 */
			sc->result = DID_SOFT_ERROR << 16;
			break;

		}	/* switch(status) */

		dprintk((KERN_NOTICE "mpiscsi: sc->result set to %08xh\n", sc->result));
	}

	if (sc != NULL) {
		unsigned long flags;

		/* Unmap the DMA buffers, if any. */
		if (sc->use_sg) {
			pci_unmap_sg(ioc->pcidev,
				     (struct scatterlist *) sc->request_buffer,
				     sc->use_sg,
				     scsi_to_pci_dma_dir(sc->sc_data_direction));
		} else if (sc->request_bufflen) {
			pci_unmap_single(ioc->pcidev,
					 (dma_addr_t)((long)sc->SCp.ptr),
					 sc->request_bufflen,
					 scsi_to_pci_dma_dir(sc->sc_data_direction));
		}

		spin_lock_irqsave(&io_request_lock, flags);
		sc->scsi_done(sc);
		spin_unlock_irqrestore(&io_request_lock, flags);
	}

	return 1;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	search_taskQ - Search SCSI task mgmt request queue for specific
 *			request type
 *	@remove: (Boolean) Should request be removed if found?
 *	@sc: Pointer to Scsi_Cmnd structure
 *	@task_type: Task type to search for
 *
 *	Returns pointer to MPI request frame if found, or %NULL if request
 *	was not found.
 */
MPI_FRAME_HDR *
search_taskQ(int remove, Scsi_Cmnd *sc, U8 task_type)
{
	MPI_FRAME_HDR *mf = NULL;
	unsigned long flags;
	int count = 0;
	int list_sz;

	dslprintk((KERN_INFO "mpiscsi: spinlock#1\n"));
	spin_lock_irqsave(&mpiscsi_taskQ_lock, flags);
	list_sz = mpiscsi_taskQ_cnt;
	if (! Q_IS_EMPTY(&mpiscsi_taskQ)) {
		mf = mpiscsi_taskQ.head;
		do {
			count++;
			if (mf->u.frame.linkage.argp1 == sc &&
			    mf->u.frame.linkage.arg1 == task_type) {
				if (remove) {
					Q_DEL_ITEM(&mf->u.frame.linkage);
					mpiscsi_taskQ_cnt--;
				}
				break;
			}
		} while ((mf = mf->u.frame.linkage.forw) != (MPI_FRAME_HDR*)&mpiscsi_taskQ);
		if (mf == (MPI_FRAME_HDR*)&mpiscsi_taskQ) {
			mf = NULL;
		}
	}
	spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);

	if (list_sz) {
		dprintk((KERN_INFO "mpiscsi: search_taskQ(%d,%p,%d) results=%p (%sFOUND%s)!\n",
				   remove, sc, task_type,
				   mf,
				   mf ? "" : "NOT_",
				   (mf && remove) ? "+REMOVED" : "" ));
		dprintk((KERN_INFO "  (searched thru %d of %d items on taskQ)\n",
				   count,
				   list_sz ));
	}

	return mf;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

/*
 *  Hack!  I'd like to report if a device is returning QUEUE_FULL
 *  but maybe not each and every time...
 */
static long last_queue_full = 0;

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	report_queue_full - Report QUEUE_FULL status returned from a SCSI
 *			target device
 *	@sc: Pointer to Scsi_Cmnd structure
 *	@pScsiReply: Pointer to SCSIIOReply_t
 *	@pScsiReq: Pointer to original SCSI request
 *
 *	This routine periodically reports QUEUE_FULL status returned from a
 *	SCSI target device.  It reports this to the console via kernel
 *	printk() API call, not more than once every 10 seconds.
 */
STATIC void
report_queue_full(Scsi_Cmnd *sc, SCSIIOReply_t *pScsiReply, SCSIIORequest_t *pScsiReq)
{
	long time = jiffies;

	if (time - last_queue_full > 10 * HZ) {
		printk(KERN_WARNING "mpi: Device reported QUEUE_FULL!  SCSI bus:target:lun = %d:%d:%d\n",
		       0, sc->target, sc->lun);
		last_queue_full = time;
	}
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
//  SCSI fops...
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_scsi_detect - Register SCSI host adapter (channels?) with linux
 *			scsi mid-layer
 *	@tpnt: Pointer to Scsi_Host_Template structure
 *
 *	(linux Scsi_Host_Template.detect routine)
 *
 *	Returns number of SCSI host adapters (channels?) were successfully
 *	registered with the linux scsi mid-layer via the scsi_register()
 *	API call.
 */
int
mpi_scsi_detect(Scsi_Host_Template *tpnt)
{
	struct Scsi_Host    *sh = NULL;
	MPI_SCSI_HOST       *hd = NULL;
	MPI_ADAPTER         *this;
	unsigned long        flags;
	int                  sz;
	U8                  *mem;

	dprintk((KERN_INFO "mpiscsi: mpi_scsi_detect()\n"));

	this = mpi_adapter_find_first();
	while (this != NULL) {
		/* FIXME!  Multi-port (aka FC929) support...
		 * for (i = 0; i < this->facts.NumberOfPorts; i++)
		 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
		tpnt->proc_dir = &proc_mpiscsi;
#endif
		sh = scsi_register(tpnt, sizeof(MPI_SCSI_HOST));
		if (sh != NULL) {
			save_flags(flags);
			cli();
			sh->io_port = 0;
			sh->n_io_port = 0;
			sh->irq = 0;

			/* Yikes!  This is important!
			 * Otherwise, by default, linux only scans target IDs 0-7!
			 */
			sh->max_id = this->pfacts0.MaxDevices - 1;

			sh->this_id = this->pfacts0.PortSCSIID;

			restore_flags(flags);

			hd = (MPI_SCSI_HOST *) sh->hostdata;
			hd->ioc = this;
			hd->port = 0;		/* FIXME! */

			/* SCSI needs Scsi_Cmnd lookup table!
			 * (with size equal to req_depth*PtrSz!)
			 */
			sz = hd->ioc->req_depth * sizeof(void *);
			mem = kmalloc(sz, GFP_KERNEL);
			if (mem == NULL)
				return mpi_scsi_hosts;

			memset(mem, 0, sz);
			hd->ScsiLookup = (struct scsi_cmnd **) mem;

			dprintk((KERN_INFO "mpiscsi: ScsiLookup @ %p, sz=%d\n",
				 hd->ScsiLookup, sz));

			/* SCSI also needs SG buckets/hunk management!
			 * (with size equal to N * req_sz * req_depth!)
			 * (where N is number of SG buckets per hunk)
			 */
			sz = MPI_SG_BUCKETS_PER_HUNK * hd->ioc->req_sz * hd->ioc->req_depth;
			mem = pci_alloc_consistent(hd->ioc->pcidev, sz,
						   &hd->SgHunksDMA);
			if (mem == NULL)
				return mpi_scsi_hosts;

			memset(mem, 0, sz);
			hd->SgHunks = (U8*)mem;

			dprintk((KERN_INFO "mpiscsi: SgHunks    @ %p(%08x), sz=%d\n",
				 hd->SgHunks, hd->SgHunksDMA, sz));

			hd->qtag_tick = jiffies;

			this->sh = sh;
			mpi_scsi_hosts++;
		}

		this = mpi_adapter_find_next(this);
	}

	return mpi_scsi_hosts;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
    static char *info_kbuf = NULL;
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_scsi_release - Unregister SCSI host from linux scsi mid-layer
 *	@host: Pointer to Scsi_Host structure
 *
 *	(linux Scsi_Host_Template.release routine)
 *	This routine releases all resources associated with the SCSI host
 *	adapter.
 *
 *	Returns 0 for success.
 */
int
mpi_scsi_release(struct Scsi_Host *host)
{
	MPI_SCSI_HOST       *hd;
	unsigned long	 flags;

// YIKES!
	spin_lock_irqsave(&mpiscsi_taskQ_lock, flags);
	if (mpiscsi_taskQ_bh_active) {
		int count = 10 * 100;

		dprintk((KERN_INFO "mpiscsi: Info: Zapping TaskMgmt thread!\n"));

		/* Zap the taskQ! */
		Q_INIT(&mpiscsi_taskQ, MPI_FRAME_HDR);
		spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);

		while(mpiscsi_taskQ_bh_active && --count) {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(1);
		}
		if (!count)
			printk(KERN_ERR "mpiscsi: ERROR! TaskMgmt thread still active!\n");
	}
	spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);

	hd = (MPI_SCSI_HOST *) host->hostdata;
	if (hd != NULL) {
		int sz1, sz2;

		sz1 = sz2 = 0;
		if (hd->ScsiLookup != NULL) {
			sz1 = hd->ioc->req_depth * sizeof(void *);
			kfree(hd->ScsiLookup);
			hd->ScsiLookup = NULL;
		}

		if (hd->SgHunks != NULL) {

			sz2 = MPI_SG_BUCKETS_PER_HUNK * hd->ioc->req_sz * hd->ioc->req_depth;
			pci_free_consistent(hd->ioc->pcidev, sz2,
					    hd->SgHunks, hd->SgHunksDMA);
			hd->SgHunks = NULL;
		}
		dprintk((KERN_INFO "mpiscsi: Free'd ScsiLookup (%d) and SgHunks (%d) memory\n", sz1, sz2));
	}

	if (mpi_scsi_hosts) {
		if (--mpi_scsi_hosts == 0) {
			;
#if 0
			flush_pending();
#endif
		}
	}

	if (info_kbuf != NULL)
		kfree(info_kbuf);

	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_scsi_info - Return information about MPI adapter
 *	@SChost: Pointer to Scsi_Host structure
 *
 *	(linux Scsi_Host_Template.info routine)
 *
 *	Returns pointer to buffer where information was written.
 */
const char *
mpi_scsi_info(struct Scsi_Host *SChost)
{
	MPI_SCSI_HOST *h;
	int size = 0;

	if (info_kbuf == NULL) {
		if ((info_kbuf = kmalloc(0x1000 /* 4Kb */, GFP_KERNEL)) == NULL) {
			return info_kbuf;
		}
	}

	h = (MPI_SCSI_HOST *)SChost->hostdata;
	info_kbuf[0] = '\0';
	mpi_print_ioc_facts(h->ioc, info_kbuf, &size, 0);
	info_kbuf[size-1] = '\0';

	return info_kbuf;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
        STATIC int max_qd = 1;
#ifdef MPI_DEBUG
        STATIC int max_sges = 0;
        STATIC int max_xfer = 0;
#endif
#if 0
	STATIC int max_num_sges = 0;
	STATIC int max_sgent_len = 0;
#endif
#if 0
STATIC int index_log[128];
STATIC int index_ent = 0;
STATIC __inline__ void ADD_INDEX_LOG(int req_ent)
{
	int i = index_ent++;

	index_log[i & (128 - 1)] = req_ent;
}
#else
#define ADD_INDEX_LOG(req_ent)	do { } while(0)
#endif
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_scsi_queuecommand - Primary SCSI IO start routine
 *	@SCpnt: Pointer to Scsi_Cmnd structure
 *	@done: Pointer SCSI mid-layer IO completion function
 *
 *	(linux Scsi_Host_Template.queuecommand routine)
 *	This is the primary SCSI IO start routine.  Create a MPI SCSIIORequest
 *	from a linux Scsi_Cmnd request and send it to the IOC.
 *
 *	Returns 0. (rtn value discarded by linux scsi mid-layer)
 */
int
mpi_scsi_queuecommand(Scsi_Cmnd *SCpnt, void (*done)(Scsi_Cmnd *))
{
	struct Scsi_Host    *host;
	MPI_SCSI_HOST       *hd;
	MPI_FRAME_HDR       *mf;
	SCSIIORequest_t     *pScsiReq;
	int                  datadir;
	U32                  len;
	U32                  sgdir;
	U32                  scsictl;
	U32                  scsidir;
	U32                  qtag;
	U32                 *mptr;
	int                  sge_spill1;
	int                  frm_sz;
	int                  sges_left;
	U32                  chain_offset;
	int                  my_idx;
	int			 i;

	dirqprintk((KERN_INFO "mpi_scsi_qcmd: SCpnt=%p, done()=%p\n",
		    SCpnt, done));

	host = SCpnt->host;
	hd = (MPI_SCSI_HOST *) host->hostdata;
#if 0
	if (host->host_busy >= 60) {
		MPI_ADAPTER *ioc = hd->ioc;
		u16 pci_command, pci_status;

		/* The IOC is probably hung, investigate status. */
		printk("MPI: IOC probably hung IOCSTAT[%08x] INTSTAT[%08x] REPLYFIFO[%08x]\n",
		       readl(&ioc->chip.fc9xx->DoorbellValue),
		       readl(&ioc->chip.fc9xx->IntStatus),
		       readl(&ioc->chip.fc9xx->ReplyFifo));
		pci_read_config_word(ioc->pcidev, PCI_COMMAND, &pci_command);
		pci_read_config_word(ioc->pcidev, PCI_STATUS, &pci_status);
		printk("MPI: PCI command[%04x] status[%04x]\n", pci_command, pci_status);
		{
			/* DUMP req index logger. */
			int begin, end;

			begin = (index_ent - 65) & (128 - 1);
			end = index_ent & (128 - 1);
			printk("MPI: REQ_INDEX_HIST[");
			while (begin != end) {
				printk("(%04x)", index_log[begin]);
				begin = (begin + 1) & (128 - 1);
			}
			printk("\n");
		}
		sti();
		while(1)
			barrier();
	}
#endif

	SCpnt->scsi_done = done;

	/* 20000617 -sralston
	 *  GRRRRR...  Shouldn't have to to this but...
	 *  Do explicit check for REQUEST_SENSE and cached SenseData.
	 *  If yes, return cached SenseData.
	 */
#ifdef MPI_SCSI_CACHE_AUTOSENSE
	{
		MPI_SCSI_DEV        *mpi_sdev;

		mpi_sdev = (MPI_SCSI_DEV *) SCpnt->device->hostdata;
		if (mpi_sdev && SCpnt->cmnd[0] == REQUEST_SENSE) {
			U8 *dest = NULL;

			if (!SCpnt->use_sg)
				dest = SCpnt->request_buffer;
			else {
				struct scatterlist *sg = (struct scatterlist *) SCpnt->request_buffer;
				if (sg)
					dest = (U8 *) (unsigned long)sg_dma_address(sg);
			}

			if (dest && mpi_sdev->sense_sz) {
				memcpy(dest, mpi_sdev->CachedSense.data, mpi_sdev->sense_sz);
#ifdef MPI_DEBUG
				{
					int  i;
					U8  *sb;

					sb = mpi_sdev->CachedSense.data;
					if (sb && ((sb[0] & 0x70) == 0x70)) {
						printk(KERN_WARNING "mpiscsi: Returning last cached SCSI (hex) SenseData:\n");
						printk(KERN_WARNING " ");
						for (i = 0; i < (8 + sb[7]); i++)
							printk("%s%02x", i == 13 ? "-" : " ", sb[i]);
						printk("\n");
					}
				}
#endif
			}
			SCpnt->resid = SCpnt->request_bufflen - mpi_sdev->sense_sz;
			SCpnt->result = 0;
//			spin_lock(&io_request_lock);
			SCpnt->scsi_done(SCpnt);
//			spin_unlock(&io_request_lock);
			return 0;
		}
	}
#endif

	if ((mf = MpiGetMsgFrame(ScsiDoneCtx, hd->ioc->id)) == NULL) {
//		SCpnt->result = DID_SOFT_ERROR << 16;
		SCpnt->result = STS_BUSY;
		SCpnt->scsi_done(SCpnt);
//		return 1;
		return 0;
	}
	pScsiReq = (SCSIIORequest_t *) mf;

	my_idx = le16_to_cpu(mf->u.frame.hwhdr.msgctxu.fld.req_idx);

	ADD_INDEX_LOG(my_idx);

	/* Map the data portion, if any. */
	sges_left = SCpnt->use_sg;
	if (sges_left) {
		sges_left = pci_map_sg(hd->ioc->pcidev,
				       (struct scatterlist *) SCpnt->request_buffer,
				       sges_left,
				       scsi_to_pci_dma_dir(SCpnt->sc_data_direction));
	} else if (SCpnt->request_bufflen) {
		dma_addr_t buf_dma_addr;

		buf_dma_addr = pci_map_single(hd->ioc->pcidev,
					      SCpnt->request_buffer,
					      SCpnt->request_bufflen,
					      scsi_to_pci_dma_dir(SCpnt->sc_data_direction));

		/* We hide it here for later unmap. */
		SCpnt->SCp.ptr = (char *)(unsigned long) buf_dma_addr;
	}

	/*
	 *  Put together a MPI SCSI request...
	 */

	/* Assume SimpleQ, NO DATA XFER for now */

	len = SCpnt->request_bufflen;
	sgdir = 0x00000000;		/* SGL IN  (host<--ioc) */
	scsidir = MPI_SCSIIO_CONTROL_NODATATRANSFER;

	/*
	 *  The scsi layer should be handling this stuff
	 *  (In 2.3.x it does -DaveM)
	 */

	/*  BUG FIX!  19991030 -sralston
	 *    TUR's being issued with scsictl=0x02000000 (DATA_IN)!
	 *    Seems we may receive a buffer (len>0) even when there
	 *    will be no data transfer!  GRRRRR...
	 */
	datadir = scsi_dir(SCpnt);
	if (datadir < 0) {
		scsidir = MPI_SCSIIO_CONTROL_READ;	/* DATA IN  (host<--ioc<--dev) */
	} else if (datadir > 0) {
		sgdir   = 0x04000000;			/* SGL OUT  (host-->ioc) */
		scsidir = MPI_SCSIIO_CONTROL_WRITE;	/* DATA OUT (host-->ioc-->dev) */
	} else {
		len = 0;
	}

	qtag = MPI_SCSIIO_CONTROL_SIMPLEQ;

	/*
	 *  Attach tags to the devices
	 */
	if (SCpnt->device->tagged_supported) {
		/*
		 *  Some drives are too stupid to handle fairness issues
		 *  with tagged queueing. We throw in the odd ordered
		 *  tag to stop them starving themselves.
		 */
		if ((jiffies - hd->qtag_tick) > (5*HZ)) {
			qtag = MPI_SCSIIO_CONTROL_ORDEREDQ;
			hd->qtag_tick = jiffies;

#if 0
			/* These are ALWAYS zero!
			 * (Because this is a place for the device driver to dynamically
			 *  assign tag numbers any way it sees fit.  That's why -DaveM)
			 */
			dprintk((KERN_DEBUG "mpiscsi: sc->device->current_tag = %08x\n",
				 SCpnt->device->current_tag));
			dprintk((KERN_DEBUG "mpiscsi: sc->tag                 = %08x\n",
				 SCpnt->tag));
#endif
		}
#if 0
		else {
			/* Hmmm...  I always see value of 0 here,
			 *  of which {HEAD_OF, ORDERED, SIMPLE} are NOT!  -sralston
			 * (Because this is a place for the device driver to dynamically
			 *  assign tag numbers any way it sees fit.  That's why -DaveM)
			 *
			 * if (SCpnt->tag == HEAD_OF_QUEUE_TAG)
			 */
			if (SCpnt->device->current_tag == HEAD_OF_QUEUE_TAG)
				qtag = MPI_SCSIIO_CONTROL_HEADOFQ;
			else if (SCpnt->tag == ORDERED_QUEUE_TAG)
				qtag = MPI_SCSIIO_CONTROL_ORDEREDQ;
		}
#endif
	}

	scsictl = scsidir | qtag;

	frm_sz = hd->ioc->req_sz;

	/* Ack!
	 * sge_spill1 = 9;
	 */
	sge_spill1 = (frm_sz - (sizeof(SCSIIORequest_t) - sizeof(SGEIOUnion_t) + sizeof(SGEChain32_t))) / 8;
	/*  spill1: for req_sz == 128 (128-48==80, 80/8==10 SGEs max, first time!), --> use 9
	 *  spill1: for req_sz ==  96 ( 96-48==48, 48/8== 6 SGEs max, first time!), --> use 5
	 */
	dsgprintk((KERN_INFO "mpiscsi: SG: %x     spill1 = %d\n",
		   my_idx, sge_spill1));

#ifdef MPI_DEBUG
	if (sges_left > max_sges) {
		max_sges = sges_left;
		dprintk((KERN_INFO "mpiscsi: MPI_MaxSges = %d\n", max_sges));
	}
#endif
#if 0
	if (sges_left > max_num_sges) {
		max_num_sges = sges_left;
		printk(KERN_INFO "mpiscsi: MPI_MaxNumSges = %d\n", max_num_sges);
	}
#endif

	dsgprintk((KERN_INFO "mpiscsi: SG: %x     sges_left = %d (initially)\n",
		   my_idx, sges_left));

	chain_offset = 0;
	if (sges_left > (sge_spill1+1)) {
#if 0
		chain_offset = 0x1E;
#endif
		chain_offset = (frm_sz - 8) / 4;
	}

	pScsiReq->TargetID = SCpnt->target;
	pScsiReq->Bus = hd->port;
	pScsiReq->ChainOffset = chain_offset;
	pScsiReq->Function = MPI_FUNCTION_SCSI_IO_REQUEST;
	pScsiReq->CDBLength = SCpnt->cmd_len;

// We have 256 bytes alloc'd per IO; let's use it.
//	pScsiReq->SenseBufferLength = SNS_LEN(SCpnt);
	pScsiReq->SenseBufferLength = 255;

	pScsiReq->Reserved = 0;
	pScsiReq->MsgFlags = 0;
	pScsiReq->LUN[0] = 0;
	pScsiReq->LUN[1] = SCpnt->lun;
	pScsiReq->LUN[2] = 0;
	pScsiReq->LUN[3] = 0;
	pScsiReq->LUN[4] = 0;
	pScsiReq->LUN[5] = 0;
	pScsiReq->LUN[6] = 0;
	pScsiReq->LUN[7] = 0;
	pScsiReq->LUN[8] = 0;
	pScsiReq->Control = cpu_to_le32(scsictl);

	/*
	 *  Write SCSI CDB into the message
	 */
	for (i = 0; i < 12; i++)
		pScsiReq->CDB[i] = SCpnt->cmnd[i];
	for (i = 12; i < 16; i++)
		pScsiReq->CDB[i] = 0;

	/* DataLength */
	pScsiReq->DataLength = cpu_to_le32(len);

	/* SenseBuffer low address */
	pScsiReq->SenseBufferLowAddr = cpu_to_le32(hd->ioc->sense_buf_pool_dma + (my_idx * 256));

	mptr = (U32 *) &pScsiReq->SGL;

	/*
	 *  Now fill in the SGList...
	 *  NOTES: For 128 byte req_sz, we can hold up to 10 simple SGE's
	 *  in the remaining request frame.  We -could- do unlimited chains
	 *  but each chain buffer can only be req_sz bytes in size, and
	 *  we lose one SGE whenever we chain.
	 *  For 128 req_sz, we can hold up to 16 SGE's per chain buffer.
	 *  For practical reasons, limit ourselves to 1 overflow chain buffer;
	 *  giving us 9 + 16 == 25 SGE's max.
	 *  At 4 Kb per SGE, that yields 100 Kb max transfer.
	 *
	 *  (This code needs to be completely changed when/if 64-bit DMA
	 *   addressing is used, since we will be able to fit much less than
	 *   10 embedded SG entries. -DaveM)
	 */
	if (sges_left) {
		struct scatterlist *sg = (struct scatterlist *) SCpnt->request_buffer;
		U32  v1, v2;
		int  sge_spill2;
		int  sge_cur_spill;
		int  sgCnt;
		U8  *pSgBucket;
		int  chain_sz;

		len = 0;

		/*        sge_spill2 = 15;
		 *  spill2: for req_sz == 128 (128/8==16 SGEs max, first time!), --> use 15
		 *  spill2: for req_sz ==  96 ( 96/8==12 SGEs max, first time!), --> use 11
		 */
		sge_spill2 = frm_sz / 8 - 1;
		dsgprintk((KERN_INFO "mpiscsi: SG: %x     spill2 = %d\n",
			   my_idx, sge_spill2));

		pSgBucket = NULL;
		sgCnt = 0;
		sge_cur_spill = sge_spill1;
		while (sges_left) {
#if 0
			if (sg_dma_len(sg) > max_sgent_len) {
				max_sgent_len = sg_dma_len(sg);
				printk(KERN_INFO "mpiscsi: MPI_MaxSgentLen = %d\n", max_sgent_len);
			}
#endif
			/* Write one simple SGE */
			v1 = sgdir | 0x10000000 | sg_dma_len(sg);
			len += sg_dma_len(sg);
			v2 = sg_dma_address(sg);
			dsgprintk((KERN_INFO "mpiscsi: SG: %x     Writing SGE @%p: %08x %08x, sges_left=%d\n",
				   my_idx, mptr, v1, v2, sges_left));
			*mptr++ = cpu_to_le32(v1);
			*mptr++ = cpu_to_le32(v2);
			sg++;
			sgCnt++;

			if (--sges_left == 0) {
				/* re-write 1st word of previous SGE with SIMPLE,
				 * LE, EOB, and EOL bits!
				 */
				v1 = 0xD1000000 | sgdir | sg_dma_len(sg-1);
				dsgprintk((KERN_INFO "mpiscsi: SG: %x (re)Writing SGE @%p: %08x (VERY LAST SGE!)\n",
					   my_idx, mptr-2, v1));
				*(mptr - 2) = cpu_to_le32(v1);
			} else {
				if ((sges_left > 1) && ((sgCnt % sge_cur_spill) == 0)) {
					dsgprintk((KERN_INFO "mpiscsi: SG: %x     SG spill at modulo 0!\n",
						   my_idx));

					/* Fixup previous SGE with LE bit! */
					v1 = sgdir | 0x90000000 | sg_dma_len(sg-1);
					dsgprintk((KERN_INFO "mpiscsi: SG: %x (re)Writing SGE @%p: %08x (LAST BUCKET SGE!)\n",
						   my_idx, mptr-2, v1));
					*(mptr - 2) = cpu_to_le32(v1);

					chain_offset = 0;
					/* Going to need another chain? */
					if (sges_left > (sge_spill2+1)) {
#if 0
						chain_offset = 0x1E;
#endif
						chain_offset = (frm_sz - 8) / 4;
						chain_sz = frm_sz;
					} else {
						chain_sz = sges_left * 8;
					}

					/* write chain SGE at mptr. */
					v1 = 0x30000000 | chain_offset<<16 | chain_sz;
					if (pSgBucket == NULL) {
						pSgBucket = hd->SgHunks
							+ (my_idx * frm_sz * MPI_SG_BUCKETS_PER_HUNK);
					} else {
						pSgBucket += frm_sz;
					}
					v2 = (hd->SgHunksDMA +
					      ((U8 *)pSgBucket - (U8 *)hd->SgHunks));
					dsgprintk((KERN_INFO "mpiscsi: SG: %x     Writing SGE @%p: %08x %08x (CHAIN!)\n",
						   my_idx, mptr, v1, v2));
					*(mptr++) = cpu_to_le32(v1);
					*(mptr) = cpu_to_le32(v2);

					mptr = (U32 *) pSgBucket;
					sgCnt = 0;
					sge_cur_spill = sge_spill2;
				}
			}
		}
	} else {
		dsgprintk((KERN_INFO "mpiscsi: SG: non-SG for %p, len=%d\n",
			   SCpnt, SCpnt->request_bufflen));

		if (len > 0) {
			dma_addr_t buf_dma_addr;

			buf_dma_addr = (dma_addr_t) (unsigned long)SCpnt->SCp.ptr;
			*(mptr++) = cpu_to_le32(0xD1000000|sgdir|SCpnt->request_bufflen);
			*(mptr++) = cpu_to_le32(buf_dma_addr);
		}
	}

#ifdef MPI_DEBUG
	/* if (SCpnt->request_bufflen > max_xfer) */
	if (len > max_xfer) {
		max_xfer = len;
		dprintk((KERN_INFO "mpiscsi: MPI_MaxXfer = %d\n", max_xfer));
	}
#endif

	hd->ScsiLookup[my_idx] = SCpnt;

	/* Main banana... */
	MpiPutMsgFrame(ScsiDoneCtx, hd->ioc->id, mf);

	atomic_inc(&queue_depth);
	if (atomic_read(&queue_depth) > max_qd) {
		max_qd = atomic_read(&queue_depth);
		dprintk((KERN_INFO "mpiscsi: Queue depth now %d.\n", max_qd));
	}

	mb();
	dirqprintk((KERN_INFO "mpiscsi: Issued SCSI cmd (%p)\n", SCpnt));

	return 0;
}

#ifdef MPI_SCSI_USE_NEW_EH		// {
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
    mpi_scsi_abort
    Returns: 0=SUCCESS, else FAILED
*/
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_scsi_abort - Abort linux Scsi_Cmnd routine, new_eh variant
 *	@SCpnt: Pointer to Scsi_Cmnd structure, IO to be aborted
 *
 *	(linux Scsi_Host_Template.eh_abort_handler routine)
 *
 *	Returns SUCCESS or FAILED.  
 */
int
mpi_scsi_abort(Scsi_Cmnd * SCpnt)
{
	MPI_FRAME_HDR       *mf;
	SCSITaskMgmt_t      *pScsiTm;
	MPI_SCSI_HOST       *hd;
	U32                 *msg;
	U32                  ctx2abort;
	int		     i;

	printk(KERN_WARNING "mpiscsi: Attempting _ABORT SCSI IO (=%p)\n", SCpnt);
	printk(KERN_WARNING "mpiscsi: IOs outstanding = %d\n", atomic_read(&queue_depth));

	hd = (MPI_SCSI_HOST *) SCpnt->host->hostdata;

	if ((mf = MpiGetMsgFrame(ScsiTaskCtx, hd->ioc->id)) == NULL) {
//		SCpnt->result = DID_SOFT_ERROR << 16;
		SCpnt->result = STS_BUSY;
		SCpnt->scsi_done(SCpnt);
		return FAILED;
	}

	pScsiTm = (SCSITaskMgmt_t *) mf;
	msg = (U32 *) mf;

	pScsiTm->TargetID = SCpnt->target;
	pScsiTm->Bus = hd->port;
	pScsiTm->ChainOffset = 0;
	pScsiTm->Function = MPI_FUNCTION_SCSI_TASK_MGMT;

	pScsiTm->Reserved = 0;
	pScsiTm->TaskType = MPI_SCSITASKMGMT_TASKTYPE_ABORT_TASK;
	pScsiTm->Reserved1 = 0;
	pScsiTm->MsgFlags = 0;

	for (i = 0; i < 8; i++) {
		U8 val = 0;
		if (i == 1)
			val = SCpnt->lun;
		pScsiTm->LUN[i] = val;
	}

	for (i = 0; i < 7; i++)
		pScsiTm->Reserved2[i] = 0;

	/* Most important!  Set TaskMsgContext to SCpnt's MsgContext!
	 * (the IO to be ABORT'd)
	 *
	 * NOTE: Since we do not byteswap MsgContext, we do not
	 *       swap it here either.  It is an opaque cookie to
	 *       the controller, so it does not matter. -DaveM
	 */
	ctx2abort = SCPNT_TO_MSGCTX(SCpnt);
	dprintk((KERN_INFO "DbG: ctx2abort = %08x\n", ctx2abort));
	pScsiTm->TaskMsgContext = ctx2abort;

	wmb();

/* MPI v0.10 requires SCSITaskMgmt requests be sent via Doorbell/handshake
	MpiPutMsgFrame(hd->ioc->id, mf);
*/
/* FIXME!  Check return status! */
	(void) SendHandShakeRequest(ScsiTaskCtx, hd->ioc->id, sizeof(SCSITaskMgmt_t), msg);

	wmb();

	return SUCCESS;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_scsi_bus_reset - Perform a SCSI BUS_RESET!  new_eh variant
 *	@SCpnt: Pointer to Scsi_Cmnd structure, IO which reset is due to
 *
 *	(linux Scsi_Host_Template.eh_bus_reset_handler routine)
 *
 *	Returns SUCCESS or FAILED.
 */
int
mpi_scsi_bus_reset(Scsi_Cmnd * SCpnt)
{
	MPI_FRAME_HDR       *mf;
	SCSITaskMgmt_t      *pScsiTm;
	MPI_SCSI_HOST       *hd;
	U32                 *msg;
	int		     i;

	printk(KERN_WARNING "mpiscsi: Attempting _BUS_RESET (%p)\n", SCpnt);
	printk(KERN_WARNING "mpiscsi: IOs outstanding = %d\n", atomic_read(&queue_depth));

	hd = (MPI_SCSI_HOST *) SCpnt->host->hostdata;

	if ((mf = MpiGetMsgFrame(ScsiTaskCtx, hd->ioc->id)) == NULL) {
//		SCpnt->result = DID_SOFT_ERROR << 16;
		SCpnt->result = STS_BUSY;
		SCpnt->scsi_done(SCpnt);
		return FAILED;
	}

	pScsiTm = (SCSITaskMgmt_t *) mf;
	msg = (U32 *) mf;

	pScsiTm->TargetID = SCpnt->target;
	pScsiTm->Bus = hd->port;
	pScsiTm->ChainOffset = 0;
	pScsiTm->Function = MPI_FUNCTION_SCSI_TASK_MGMT;

	pScsiTm->Reserved = 0;
	pScsiTm->TaskType = MPI_SCSITASKMGMT_TASKTYPE_RESET_BUS;
	pScsiTm->Reserved1 = 0;
	pScsiTm->MsgFlags = 0;

	for (i = 0; i < 8; i++)
		pScsiTm->LUN[i] = 0;

	/* Control: No data direction, set task mgmt bit? */
	for (i = 0; i < 7; i++)
		pScsiTm->Reserved2[i] = 0;

	pScsiTm->TaskMsgContext = cpu_to_le32(0);

	wmb();

/* MPI v0.10 requires SCSITaskMgmt requests be sent via Doorbell/handshake
	MpiPutMsgFrame(hd->ioc->id, mf);
*/
/* FIXME!  Check return status! */
	(void) SendHandShakeRequest(ScsiTaskCtx, hd->ioc->id, sizeof(SCSITaskMgmt_t), msg);

	wmb();

	return SUCCESS;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_scsi_dev_reset - Perform a SCSI TARGET_RESET!  new_eh variant
 *	@SCpnt: Pointer to Scsi_Cmnd structure, IO which reset is due to
 *
 *	(linux Scsi_Host_Template.eh_dev_reset_handler routine)
 *
 *	Returns SUCCESS or FAILED.
 */
int
mpi_scsi_dev_reset(Scsi_Cmnd * SCpnt)
{
	MPI_FRAME_HDR       *mf;
	SCSITaskMgmt_t      *pScsiTm;
	MPI_SCSI_HOST       *hd;
	U32                 *msg;
	int                  i;

	printk(KERN_WARNING "mpiscsi: Attempting _TARGET_RESET (%p)\n", SCpnt);
	printk(KERN_WARNING "mpiscsi: IOs outstanding = %d\n", atomic_read(&queue_depth));

	hd = (MPI_SCSI_HOST *) SCpnt->host->hostdata;

	if ((mf = MpiGetMsgFrame(ScsiTaskCtx, hd->ioc->id)) == NULL) {
//		SCpnt->result = DID_SOFT_ERROR << 16;
		SCpnt->result = STS_BUSY;
		SCpnt->scsi_done(SCpnt);
		return FAILED;
	}

	pScsiTm = (SCSITaskMgmt_t *) mf;
	msg = (U32*)mf;

	pScsiTm->TargetID = SCpnt->target;
	pScsiTm->Bus = hd->port;
	pScsiTm->ChainOffset = 0;
	pScsiTm->Function = MPI_FUNCTION_SCSI_TASK_MGMT;

	pScsiTm->Reserved = 0;
	pScsiTm->TaskType = MPI_SCSITASKMGMT_TASKTYPE_TARGET_RESET;
	pScsiTm->Reserved1 = 0;
	pScsiTm->MsgFlags = 0;

	/* _TARGET_RESET goes to LUN 0 always! */
	for (i = 0; i < 8; i++)
		pScsiTm->LUN[i] = 0;

	/* Control: No data direction, set task mgmt bit? */
	for (i = 0; i < 7; i++)
		pScsiTm->Reserved2[i] = 0;

	pScsiTm->TaskMsgContext = cpu_to_le32(0);

	wmb();

/* MPI v0.10 requires SCSITaskMgmt requests be sent via Doorbell/handshake
	MpiPutMsgFrame(hd->ioc->id, mf);
*/
/* FIXME!  Check return status! */
	(void) SendHandShakeRequest(ScsiTaskCtx, hd->ioc->id, sizeof(SCSITaskMgmt_t), msg);

	wmb();

	return SUCCESS;
}

#if 0	// {
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_scsi_host_reset - Perform a SCSI host adapter RESET!
 *			new_eh variant
 *	@SCpnt: Pointer to Scsi_Cmnd structure, IO which reset is due to
 *
 *	(linux Scsi_Host_Template.eh_host_reset_handler routine)
 *
 *	Returns SUCCESS or FAILED.
 */
int
mpi_scsi_host_reset(Scsi_Cmnd * SCpnt)
{
	return FAILED;
}
#endif	// }

#else		// MPI_SCSI old EH stuff...
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_scsi_old_abort - Abort linux Scsi_Cmnd routine
 *	@SCpnt: Pointer to Scsi_Cmnd structure, IO to be aborted
 *
 *	(linux Scsi_Host_Template.abort routine)
 *
 *	Returns SCSI_ABORT_{SUCCESS,BUSY,PENDING}.
 */
int
mpi_scsi_old_abort(Scsi_Cmnd *SCpnt)
{
	MPI_SCSI_HOST		*hd;
	MPI_FRAME_HDR		*mf;
	struct tq_struct	*ptaskfoo;
	unsigned long		 flags;

	printk(KERN_WARNING "mpiscsi: Scheduling _ABORT SCSI IO (=%p)\n", SCpnt);
	printk(KERN_WARNING "mpiscsi: IOs outstanding = %d\n", atomic_read(&queue_depth));

	if ((hd = (MPI_SCSI_HOST *) SCpnt->host->hostdata) == NULL) {
		SCpnt->result = DID_ABORT << 16;
		SCpnt->scsi_done(SCpnt);
		return SCSI_ABORT_SUCCESS;
	}

	/*
	 *  Check to see if there's already an ABORT queued for this guy.
	 */
	mf = search_taskQ(0,SCpnt,MPI_SCSITASKMGMT_TASKTYPE_ABORT_TASK);
	if (mf != NULL) {
		return SCSI_ABORT_PENDING;
	}

	if ((mf = MpiGetMsgFrame(ScsiTaskCtx, hd->ioc->id)) == NULL) {
//		SCpnt->result = DID_SOFT_ERROR << 16;
		SCpnt->result = STS_BUSY;
		SCpnt->scsi_done(SCpnt);
		return SCSI_ABORT_BUSY;
	}

	/*
	 *  Add ourselves to (end of) mpiscsi_taskQ.
	 *  Check to see if our _bh is running.  If NOT, schedule it.
	 */
	dslprintk((KERN_INFO "mpiscsi: spinlock#2\n"));
	spin_lock_irqsave(&mpiscsi_taskQ_lock, flags);
	Q_ADD_TAIL(&mpiscsi_taskQ, &mf->u.frame.linkage, MPI_FRAME_HDR);
	mpiscsi_taskQ_cnt++;
	/* Yikes - linkage! */
//	SCpnt->host_scribble = (unsigned char *)mf;
	mf->u.frame.linkage.arg1 = MPI_SCSITASKMGMT_TASKTYPE_ABORT_TASK;
	mf->u.frame.linkage.argp1 = SCpnt;
	if (! mpiscsi_taskQ_bh_active) {
		mpiscsi_taskQ_bh_active = 1;
		/*
		 *  Oh how cute, no alloc/free/mgmt needed if we use
		 *  (bottom/unused portion of) MPI request frame.
		 */
		ptaskfoo = (struct tq_struct *) ((U8*)mf + hd->ioc->req_sz - sizeof(*ptaskfoo));
		ptaskfoo->next = NULL;
		ptaskfoo->sync = 0;
		ptaskfoo->routine = mpi_scsi_taskmgmt_bh;
		ptaskfoo->data = SCpnt;

		queue_task(ptaskfoo, &tq_scheduler);
	}
	spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);

	return SCSI_ABORT_PENDING;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpi_scsi_old_reset - Perform a SCSI BUS_RESET!
 *	@SCpnt: Pointer to Scsi_Cmnd structure, IO which reset is due to
 *	@reset_flags: (not used?)
 *
 *	(linux Scsi_Host_Template.reset routine)
 *
 *	Returns SCSI_RESET_{SUCCESS,PUNT,PENDING}.
 */
int
mpi_scsi_old_reset(Scsi_Cmnd *SCpnt, unsigned int reset_flags)
{
	MPI_SCSI_HOST		*hd;
	MPI_FRAME_HDR		*mf;
	struct tq_struct	*ptaskfoo;
	unsigned long		 flags;

	printk(KERN_WARNING "mpiscsi: Scheduling _BUS_RESET (=%p)\n", SCpnt);
	printk(KERN_WARNING "mpiscsi: IOs outstanding = %d\n", atomic_read(&queue_depth));

	if ((hd = (MPI_SCSI_HOST *) SCpnt->host->hostdata) == NULL) {
		SCpnt->result = DID_RESET << 16;
		SCpnt->scsi_done(SCpnt);
		return SCSI_RESET_SUCCESS;
	}

	/*
	 *  Check to see if there's already a BUS_RESET queued for this guy.
	 */
	mf = search_taskQ(0,SCpnt,MPI_SCSITASKMGMT_TASKTYPE_RESET_BUS);
	if (mf != NULL) {
		return SCSI_RESET_PENDING;
	}

	if ((mf = MpiGetMsgFrame(ScsiTaskCtx, hd->ioc->id)) == NULL) {
//		SCpnt->result = DID_SOFT_ERROR << 16;
		SCpnt->result = STS_BUSY;
		SCpnt->scsi_done(SCpnt);
		return SCSI_RESET_PUNT;
	}

	/*
	 *  Add ourselves to (end of) mpiscsi_taskQ.
	 *  Check to see if our _bh is running.  If NOT, schedule it.
	 */
	dslprintk((KERN_INFO "mpiscsi: spinlock#3\n"));
	spin_lock_irqsave(&mpiscsi_taskQ_lock, flags);
	Q_ADD_TAIL(&mpiscsi_taskQ, &mf->u.frame.linkage, MPI_FRAME_HDR);
	mpiscsi_taskQ_cnt++;
	/* Yikes - linkage! */
//	SCpnt->host_scribble = (unsigned char *)mf;
	mf->u.frame.linkage.arg1 = MPI_SCSITASKMGMT_TASKTYPE_RESET_BUS;
	mf->u.frame.linkage.argp1 = SCpnt;
	if (! mpiscsi_taskQ_bh_active) {
		mpiscsi_taskQ_bh_active = 1;
		/*
		 *  Oh how cute, no alloc/free/mgmt needed if we use
		 *  (bottom/unused portion of) MPI request frame.
		 */
		ptaskfoo = (struct tq_struct *) ((U8*)mf + hd->ioc->req_sz - sizeof(*ptaskfoo));
		ptaskfoo->next = NULL;
		ptaskfoo->sync = 0;
		ptaskfoo->routine = mpi_scsi_taskmgmt_bh;
		ptaskfoo->data = SCpnt;

		queue_task(ptaskfoo, &tq_scheduler);
	}
	spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);

	return SCSI_RESET_PENDING;
}

#endif		// }

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	mpi_scsi_taskmgmt_bh - SCSI task mgmt bottom half handler
 *	@sc: (unused)
 *
 *	This routine (thread) is active whenever there are any outstanding
 *	SCSI task management requests for a SCSI host adapter.
 *	IMPORTANT!  This routine is scheduled therefore should never be
 *	running in ISR context.  i.e., it's safe to sleep here.
 */
void
mpi_scsi_taskmgmt_bh(void *sc)
{
	Scsi_Cmnd	*SCpnt;
	MPI_FRAME_HDR	*mf;
	SCSITaskMgmt_t	*pScsiTm;
	MPI_SCSI_HOST	*hd;
	U32		 ctx2abort = 0;
	int		 i;
	unsigned long	 flags;
	U8		 task_type;

	dslprintk((KERN_INFO "mpiscsi: spinlock#4\n"));
	spin_lock_irqsave(&mpiscsi_taskQ_lock, flags);
	mpiscsi_taskQ_bh_active = 1;
	spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);

	while (1) {
		/*
		 *  We MUST remove item from taskQ *before* we format the
		 *  frame as a SCSITaskMgmt request and send it down to the IOC.
		 */
		dslprintk((KERN_INFO "mpiscsi: spinlock#5\n"));
		spin_lock_irqsave(&mpiscsi_taskQ_lock, flags);
		if (Q_IS_EMPTY(&mpiscsi_taskQ)) {
			spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);
			break;
		}
		mf = mpiscsi_taskQ.head;
		Q_DEL_ITEM(&mf->u.frame.linkage);
		mpiscsi_taskQ_cnt--;
		mpiscsi_active_taskmgmt_mf = mf;
		spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);

		SCpnt = (Scsi_Cmnd*)mf->u.frame.linkage.argp1;
		if (SCpnt == NULL) {
			printk(KERN_ERR "mpiscsi: ERROR: TaskMgmt has NULL SCpnt! (%p:%p)\n", mf, SCpnt);
			continue;
		}
		pScsiTm = (SCSITaskMgmt_t *) mf;

		for (i = 0; i < 8; i++) {
			pScsiTm->LUN[i] = 0;
		}

		task_type = mf->u.frame.linkage.arg1;
		if (task_type == MPI_SCSITASKMGMT_TASKTYPE_ABORT_TASK)
		{
			printk(KERN_WARNING "mpiscsi: Attempting _ABORT SCSI IO! (mf:sc=%p:%p)\n", mf, SCpnt);

			/* Most important!  Set TaskMsgContext to SCpnt's MsgContext!
			 * (the IO to be ABORT'd)
			 *
			 * NOTE: Since we do not byteswap MsgContext, we do not
			 *       swap it here either.  It is an opaque cookie to
			 *       the controller, so it does not matter. -DaveM
			 */
			ctx2abort = SCPNT_TO_MSGCTX(SCpnt);
			pScsiTm->LUN[1] = SCpnt->lun;
		}
		else if (task_type == MPI_SCSITASKMGMT_TASKTYPE_RESET_BUS)
		{
			printk(KERN_WARNING "mpiscsi: Attempting _BUS_RESET! (against SCSI IO mf:sc=%p:%p)\n", mf, SCpnt);
		}
		//else if (task_type == MPI_SCSITASKMGMT_TASKTYPE_TARGET_RESET) {}
		//else if (task_type == MPI_SCSITASKMGMT_TASKTYPE_ABRT_TASK_SET) {}

		printk(KERN_WARNING "mpiscsi: IOs outstanding = %d\n", atomic_read(&queue_depth));

		hd = (MPI_SCSI_HOST *) SCpnt->host->hostdata;

		pScsiTm->TargetID = SCpnt->target;
		pScsiTm->Bus = hd->port;
		pScsiTm->ChainOffset = 0;
		pScsiTm->Function = MPI_FUNCTION_SCSI_TASK_MGMT;

		pScsiTm->Reserved = 0;
		pScsiTm->TaskType = task_type;
		pScsiTm->Reserved1 = 0;
		pScsiTm->MsgFlags = 0;

		for (i = 0; i < 7; i++)
			pScsiTm->Reserved2[i] = 0;

		dprintk((KERN_INFO "DbG: ctx2abort = %08x\n", ctx2abort));
		pScsiTm->TaskMsgContext = ctx2abort;

		/* Control: No data direction, set task mgmt bit? */

		/*
		 *  As of MPI v0.10 this request can NOT be sent (normally)
		 *  via FIFOs.  So we can't:
		 *		MpiPutMsgFrame(ScsiTaskCtx, hd->ioc->id, mf);
		 *  SCSITaskMgmt requests MUST be sent ONLY via
		 *  Doorbell/handshake now.   :-(
		 *
		 *  FIXME!  Check return status!
		 */
		(void) SendHandShakeRequest(ScsiTaskCtx, hd->ioc->id, sizeof(SCSITaskMgmt_t), (U32*)mf);

		/* Spin-Wait for TaskMgmt complete!!! */
		while (mpiscsi_active_taskmgmt_mf != NULL) {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(HZ/2);
		}
	}

	dslprintk((KERN_INFO "mpiscsi: spinlock#6\n"));
	spin_lock_irqsave(&mpiscsi_taskQ_lock, flags);
	mpiscsi_taskQ_bh_active = 0;
	spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);

	return;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	ScsiTaskMgmtComplete - Callback routine registered to mpi (base)
 *			driver
 *	@ioc: Pointer to MPI_ADAPTER structure
 *	@mf: Pointer to SCSI task mgmt request frame
 *	@r: Pointer to SCSI task mgmt reply frame
 *
 *	This routine is called from mpi.c::mpi_interrupt() at the completion
 *	of any SCSI task management request.
 *	This routine is registered with the mpi (base) driver at driver
 *	load/init time via the mpi_register() API call.
 *
 *	Returns 1 indicating alloc'd request frame ptr should be freed.
 */
static int
ScsiTaskMgmtComplete(MPI_ADAPTER *ioc, MPI_FRAME_HDR *mf, MPI_FRAME_HDR *r)
{
	SCSITaskMgmtReply_t *pScsiTmReply;
	SCSITaskMgmt_t      *pScsiTmReq;
	U8                   tmType;
	unsigned long		 flags;

	dprintk((KERN_INFO "mpiscsi: SCSI TaskMgmt completed mf=%p, r=%p\n",
		 mf, r));

	dslprintk((KERN_INFO "mpiscsi: spinlock#7\n"));
	spin_lock_irqsave(&mpiscsi_taskQ_lock, flags);
	// It better be the active one!
	if (mf != mpiscsi_active_taskmgmt_mf) {
		printk(KERN_ERR "mpiscsi: ERROR! Non-active TaskMgmt (=%p) completed!\n", mf);
		mpiscsi_active_taskmgmt_mf = NULL;
		spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);
		return 1;
	}

#ifdef MPI_DEBUG
	if ((mf == NULL) ||
	    (mf >= MPI_INDEX_2_MFPTR(ioc, ioc->req_depth))) {
		printk(KERN_ERR "mpiscsi: ERROR! NULL or BAD TaskMgmt ptr (=%p)!\n", mf);
		mpiscsi_active_taskmgmt_mf = NULL;
		spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);
		return 1;
	}
#endif
	spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);

	if (r != NULL) {
		pScsiTmReply = (SCSITaskMgmtReply_t*)r;
		pScsiTmReq = (SCSITaskMgmt_t*)mf;

		/* Figure out if this was ABORT_TASK, TARGET_RESET, or BUS_RESET! */
		tmType = pScsiTmReq->TaskType;

		dprintk((KERN_INFO "mpiscsi: TaskType = %d\n", tmType));
		dprintk((KERN_INFO "mpiscsi: TerminationCount = %d\n",
			 le32_to_cpu(pScsiTmReply->TerminationCount)));

		/* Error?  (anything non-zero?) */
		if (*(U32 *)&pScsiTmReply->Reserved2[0]) {
			dprintk((KERN_INFO "mpiscsi: SCSI TaskMgmt (%d) - Oops!\n", tmType));
			dprintk((KERN_INFO "mpiscsi: IOCStatus = %04xh\n",
				 le16_to_cpu(pScsiTmReply->IOCStatus)));
			dprintk((KERN_INFO "mpiscsi: IOCLogInfo = %08xh\n",
				 le32_to_cpu(pScsiTmReply->IOCLogInfo)));
		} else {
			dprintk((KERN_INFO "mpiscsi: SCSI TaskMgmt (%d) SUCCESS!\n", tmType));
		}
	}

	/*
	 *  Signal to _bh thread that we finished.
	 */
	dslprintk((KERN_INFO "mpiscsi: spinlock#8\n"));
	spin_lock_irqsave(&mpiscsi_taskQ_lock, flags);
	mpiscsi_active_taskmgmt_mf = NULL;
	spin_unlock_irqrestore(&mpiscsi_taskQ_lock, flags);

	return 1;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *	This is anyones guess quite frankly.
 */

int
mpi_scsi_bios_param(Disk * disk, kdev_t dev, int *ip)
{
	int size;

	size = disk->capacity;
	ip[0] = 64;		/* heads                        */
	ip[1] = 32;		/* sectors                      */
	if ((ip[2] = size >> 11) > 1024) {	/* cylinders, test for big disk */
		ip[0] = 255;	/* heads                        */
		ip[1] = 63;	/* sectors                      */
		ip[2] = size / (255 * 63);	/* cylinders                    */
	}
	return 0;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Private routines...
 */
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/* 19991030 -sralston
 *  Return absolute SCSI data direction:
 *     1 = _DATA_OUT
 *     0 = _DIR_NONE
 *    -1 = _DATA_IN
 */
int
scsi_dir(Scsi_Cmnd *cmd)
{
    switch (cmd->cmnd[0]) {
        /*  _DATA_OUT commands  */
        case WRITE_6:           case WRITE_10:          case WRITE_12:
        case WRITE_LONG:        case WRITE_SAME:        case WRITE_BUFFER:
        case WRITE_VERIFY:      case WRITE_VERIFY_12:
        case COMPARE:           case COPY:              case COPY_VERIFY:
        case SEARCH_EQUAL:      case SEARCH_HIGH:       case SEARCH_LOW:
        case SEARCH_EQUAL_12:   case SEARCH_HIGH_12:    case SEARCH_LOW_12:
        case MODE_SELECT:       case MODE_SELECT_10:    case LOG_SELECT:
        case SEND_DIAGNOSTIC:   case CHANGE_DEFINITION: case UPDATE_BLOCK:
        case SET_WINDOW:        case MEDIUM_SCAN:       case SEND_VOLUME_TAG:
        case REASSIGN_BLOCKS:
        case PERSISTENT_RESERVE_OUT:
        case 0xea:
            return 1;

        /*  No data transfer commands  */
        case SEEK_6:            case SEEK_10:
        case RESERVE:           case RELEASE:
        case TEST_UNIT_READY:
        case START_STOP:
        case ALLOW_MEDIUM_REMOVAL:
            return 0;

        /*  Conditional data transfer commands  */
        case FORMAT_UNIT:
            if (cmd->cmnd[1] & 0x10)        // FmtData (data out phase)?
                return 1;
            else
                return 0;

        case VERIFY:
            if (cmd->cmnd[1] & 0x02)        // VERIFY:BYTCHK (data out phase)?
                return 1;
            else
                return 0;

        case RESERVE_10:
            if (cmd->cmnd[1] & 0x03)        // RESERSE:{LongID|Extent} (data out phase)?
                return 1;
            else
                return 0;

/* ??
        case REZERO_UNIT:       // (or REWIND)
        case SPACE:
        case ERASE:             case ERASE_10:
        case SYNCHRONIZE_CACHE:
        case LOCK_UNLOCK_CACHE:
*/

        /*  Must be data _IN!  */
        default:
            return -1;
    }
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static void
copy_sense_data(Scsi_Cmnd *sc, MPI_SCSI_HOST *hd, MPI_FRAME_HDR *mf, SCSIIOReply_t *pScsiReply)
{
	MPI_SCSI_DEV *mpi_sdev = NULL;
	U32 sense_count = le32_to_cpu(pScsiReply->SenseCount);

	if (sc && sc->device)
		mpi_sdev = (MPI_SCSI_DEV*) sc->device->hostdata;

	if (sense_count) {
		U8 *sense_data;
		int req_index;

		/* Copy the sense received into the scsi command block. */
		req_index = le16_to_cpu(mf->u.frame.hwhdr.msgctxu.fld.req_idx);
		sense_data = ((U8 *)hd->ioc->sense_buf_pool + (req_index * 256));
		memcpy(sc->sense_buffer, sense_data, SNS_LEN(sc));
		/* Cache SenseData for this SCSI device! */
		if (mpi_sdev) {
			memcpy(mpi_sdev->CachedSense.data, sense_data, sense_count);
			mpi_sdev->sense_sz = sense_count;
		}
	} else {
		dprintk((KERN_INFO "mpiscsi: Hmmm... SenseData len=0! (?)\n"));
	}

// EXP!
{
	char devFoo[32];
	IO_Info_t thisIo;

	thisIo.cdbPtr = sc->cmnd;
	thisIo.sensePtr = sc->sense_buffer;
	thisIo.SCSIStatus = pScsiReply->SCSIStatus;
	thisIo.DoDisplay = 1;
	sprintf(devFoo, "ioc%d,scsi%d:%d", hd->ioc->id, sc->target, sc->lun);
	thisIo.DevIDStr = devFoo;
// fubar
	thisIo.dataPtr = NULL;
	thisIo.inqPtr = NULL;
	if (sc->device) {
		thisIo.inqPtr = sc->device->vendor-8;		// FIXME!!!
	}

	(void) ScsiErrorReport(&thisIo);
}

	return;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
static U32
SCPNT_TO_MSGCTX(Scsi_Cmnd *sc)
{
	MPI_SCSI_HOST *hd;
	MPI_FRAME_HDR *mf;
	int i;

	hd = (MPI_SCSI_HOST *) sc->host->hostdata;

	for (i = 0; i < hd->ioc->req_depth; i++) {
		if (hd->ScsiLookup[i] == sc) {
			mf = MPI_INDEX_2_MFPTR(hd->ioc, i);
			return mf->u.frame.hwhdr.msgctxu.MsgContext;
		}
	}

	return -1;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/* symsense.c stuff... */
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 * Private data...
 * To protect ourselves against those that would pass us bogus pointers
 */
static u8 dummyInqData[SCSI_STD_INQUIRY_BYTES]
  = { 0x1F, 0x00, 0x00, 0x00,
      0x1F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
static u8 dummySenseData[SCSI_STD_SENSE_BYTES]
  = { 0x70, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0A,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00 };
static u8 dummyCDB[16]
  = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
static u8 dummyScsiData[16]
  = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

#if 0
static const char *PeripheralDeviceTypeString[32] = {
	"Direct-access",		/* 00h */
	"Sequential-access",		/* 01h */
	"Printer",			/* 02h */
	"Processor",			/* 03h */
	//	"Write-Once-Read-Multiple",	/* 04h */
	"WORM",				/* 04h */
	"CD-ROM",			/* 05h */
	"Scanner",			/* 06h */
	"Optical memory",		/* 07h */
	"Media Changer",		/* 08h */
	"Communications",		/* 09h */
	"(Graphics arts pre-press)",	/* 0Ah */
	"(Graphics arts pre-press)",	/* 0Bh */
	"Array controller",		/* 0Ch */
	"Enclosure services",		/* 0Dh */
	"Simplified direct-access",	/* 0Eh */
	"Reserved-0Fh",			/* 0Fh */
	"Reserved-10h",			/* 10h */
	"Reserved-11h",			/* 11h */
	"Reserved-12h",			/* 12h */
	"Reserved-13h",			/* 13h */
	"Reserved-14h",			/* 14h */
	"Reserved-15h",			/* 15h */
	"Reserved-16h",			/* 16h */
	"Reserved-17h",			/* 17h */
	"Reserved-18h",			/* 18h */
	"Reserved-19h",			/* 19h */
	"Reserved-1Ah",			/* 1Ah */
	"Reserved-1Bh",			/* 1Bh */
	"Reserved-1Ch",			/* 1Ch */
	"Reserved-1Dh",			/* 1Dh */
	"Reserved-1Eh",			/* 1Eh */
	"Unknown"			/* 1Fh */
};
#endif

static char *ScsiStatusString[49] = {
        "GOOD",					/* 00h */
	"Bad-Reserved-01h",			/* 01h */
        "CHECK CONDITION",			/* 02h */
	"Bad-Reserved-03h",			/* 03h */
        "CONDITION MET",			/* 04h */
	"Bad-Reserved-05h",			/* 05h */
	"Bad-Reserved-06h",			/* 06h */
	"Bad-Reserved-07h",			/* 07h */
        "BUSY",					/* 08h */
	"Bad-Reserved-09h",			/* 09h */
	"Bad-Reserved-0Ah",			/* 0Ah */
	"Bad-Reserved-0Bh",			/* 0Bh */
	"Bad-Reserved-0Ch",			/* 0Ch */
	"Bad-Reserved-0Dh",			/* 0Dh */
	"Bad-Reserved-0Eh",			/* 0Eh */
	"Bad-Reserved-0Fh",			/* 0Fh */
        "INTERMEDIATE",				/* 10h */
	"Bad-Reserved-11h",			/* 11h */
	"Bad-Reserved-12h",			/* 12h */
	"Bad-Reserved-13h",			/* 13h */
        "INTERMEDIATE-CONDITION MET",		/* 14h */
	"Bad-Reserved-15h",			/* 15h */
	"Bad-Reserved-16h",			/* 16h */
	"Bad-Reserved-17h",			/* 17h */
        "RESERVATION CONFLICT",			/* 18h */
	"Bad-Reserved-19h",			/* 19h */
	"Bad-Reserved-1Ah",			/* 1Ah */
	"Bad-Reserved-1Bh",			/* 1Bh */
	"Bad-Reserved-1Ch",			/* 1Ch */
	"Bad-Reserved-1Dh",			/* 1Dh */
	"Bad-Reserved-1Eh",			/* 1Eh */
	"Bad-Reserved-1Fh",			/* 1Fh */
	"Bad-Reserved-20h",			/* 20h */
	"Bad-Reserved-21h",			/* 21h */
        "COMMAND TERMINATED",			/* 22h */
	"Bad-Reserved-23h",			/* 23h */
	"Bad-Reserved-24h",			/* 24h */
	"Bad-Reserved-25h",			/* 25h */
	"Bad-Reserved-26h",			/* 26h */
	"Bad-Reserved-27h",			/* 27h */
        "TASK SET FULL",			/* 28h */
	"Bad-Reserved-29h",			/* 29h */
	"Bad-Reserved-2Ah",			/* 2Ah */
	"Bad-Reserved-2Bh",			/* 2Bh */
	"Bad-Reserved-2Ch",			/* 2Ch */
	"Bad-Reserved-2Dh",			/* 2Dh */
	"Bad-Reserved-2Eh",			/* 2Eh */
	"Bad-Reserved-2Fh",			/* 2Fh */
	"ACA ACTIVE"				/* 30h */
};

static const char *SenseKeyString[] = {
	"NO SENSE",			/* 0h */
	"RECOVERED ERROR",		/* 1h */
	"NOT READY",			/* 2h */
	"MEDIUM ERROR",			/* 3h */
	"HARDWARE ERROR",		/* 4h */
	"ILLEGAL REQUEST",		/* 5h */
	"UNIT ATTENTION",		/* 6h */
	"DATA PROTECT",			/* 7h */
	"BLANK CHECK",			/* 8h */
	"VENDOR-SPECIFIC",		/* 9h */
	"ABORTED COPY",			/* Ah */
	"ABORTED COMMAND",		/* Bh */
	"EQUAL (obsolete)",		/* Ch */
	"VOLUME OVERFLOW",		/* Dh */
	"MISCOMPARE",			/* Eh */
	"RESERVED",			/* Fh */
};

#if 0
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  Sense_Key_Specific() - If Sense_Key_Specific_Valid bit is set,
 *                         then print additional information via
 *                         a call to SDMS_SystemAlert().
 *
 *  Return: nothing
 */
static void
Sense_Key_Specific(IO_Info_t *ioop, char *msg1)
{
    u8 *sd;
    u8  BadValue;
    u8  SenseKey;
    int    Offset;
	int len = strlen(msg1);

    sd = ioop->sensePtr;
    if (SD_Additional_Sense_Length(sd) < 8)
        return;

    SenseKey = SD_Sense_Key(sd);

    if (SD_Sense_Key_Specific_Valid(sd))
    {
        if (SenseKey == SK_ILLEGAL_REQUEST)
        {
            Offset = SD_Bad_Byte(sd);
            if (SD_Was_Illegal_Request(sd))
            {
                BadValue = ioop->cdbPtr[Offset];
                len += sprintf(msg1+len, "\n  Illegal CDB value=%02Xh found at CDB ",
                    BadValue);
            }
            else
            {
                BadValue = ioop->dataPtr[Offset];
                len += sprintf(msg1+len, "\n  Illegal DATA value=%02Xh found at DATA ",
                    BadValue);
            }
            len += sprintf(msg1+len, "byte=%02Xh", Offset);
            if (SD_SKS_Bit_Pointer_Valid(sd))
                len += sprintf(msg1+len, "/bit=%1Xh", SD_SKS_Bit_Pointer(sd));
        }
        else if (
            (SenseKey == SK_RECOVERED_ERROR) ||
            (SenseKey == SK_HARDWARE_ERROR) ||
            (SenseKey == SK_MEDIUM_ERROR)
          )
        {
            len += sprintf(msg1+len, "\n  Recovery algorithm Actual_Retry_Count=%02Xh",
                SD_Actual_Retry_Count(sd));
        }
    }
}
#endif

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*
 *  SCSI Error Report; desired output format...
 *---
SCSI Error Report =-=-=-=-=-=-=-=-=-=-=-=-=-= (ioc0,scsi0:0)
  SCSI_Status=02h (CHECK CONDITION)
  Original_CDB[]: 00 00 00 00 00 00 - TestUnitReady
  SenseData[12h]: 70 00 06 00 00 00 00 0A 00 00 00 00 29 00 03 00 00 00
  SenseKey=6h (UNIT ATTENTION); FRU=03h
  ASC/ASCQ=29h/00h, "POWER ON, RESET, OR BUS DEVICE RESET OCCURRED"
 *---
 */

#define SPECIAL_ASCQ(c,q) \
	(((c) == 0x40 && (q) != 0x00) || ((c) == 0x4D) || ((c) == 0x70))

int
dump_cdb(char *foo, unsigned char *cdb)
{
	int i, grpCode, cdbLen;
	int l = 0;

        grpCode = cdb[0] >> 5;
        if (grpCode < 1)
            cdbLen = 6;
        else if (grpCode < 3)
            cdbLen = 10;
        else if (grpCode == 5)
            cdbLen = 12;
        else
            cdbLen = 16;

        for (i=0; i < cdbLen; i++)
            l += sprintf(foo+l, " %02X", cdb[i]);

	return l;
}

int
dump_sd(char *foo, unsigned char *sd)
{
	int snsLen = 8 + SD_Additional_Sense_Length(sd);
	int l = 0;
	int i;

        for (i=0; i < MIN(snsLen,18); i++)
            l += sprintf(foo+l, " %02X", sd[i]);
	l += sprintf(foo+l, "%s", snsLen>18 ? " ..." : "");

	return l;
}

/*  Do ASC/ASCQ lookup to english readable string  */
char *
ascq_lookup(u8 ASC, u8 ASCQ)
{
	int	 idx;
	char	*ascqStr = NULL;

	if (ASCQ_TablePtr == NULL)
		return NULL;

	/*
	 *  Need to check ASC here; if it is "special," then
	 *  the ASCQ is variable, and indicates failed component number.
	 *  We must treat the ASCQ as a "don't care" while searching the
	 *  ASCQ_Table[] by masking it off, and then restoring it later
	 *  on when we actually need to identify the failed component.
	 */
	if (SPECIAL_ASCQ(ASC,ASCQ))
		ASCQ = 0xFF;

	/*  OK, now search ASCQ_Table[] for a matching entry  */
	for (idx = 0; ASCQ_TablePtr && idx < ASCQ_TableSz; idx++)
		if ((ASC == ASCQ_TablePtr[idx].ASC) && (ASCQ == ASCQ_TablePtr[idx].ASCQ))
			return (ascqStr = ASCQ_TablePtr[idx].Description);

	if ((ASC >= 0x80) || (ASCQ >= 0x80))
		ascqStr = "(Vendor Unique)";
	else
		ascqStr = "(no matching ASC/ASCQ description found)";

	return ascqStr;
}

int
ScsiErrorReport(IO_Info_t *ioop)
{
    	char		 foo[512];
	int		 sk		= SD_Sense_Key(ioop->sensePtr);
	const char	*skstr		= SenseKeyString[sk];
	unsigned char	 asc		= SD_ASC(ioop->sensePtr);
	unsigned char	 ascq		= SD_ASCQ(ioop->sensePtr);
	const char	*ascqStr;
	int		 l;

	/*
	 *  Protect ourselves...
	 */
	if (ioop->cdbPtr == NULL)
		ioop->cdbPtr = dummyCDB;
	if (ioop->sensePtr == NULL)
		ioop->sensePtr = dummySenseData;
	if (ioop->inqPtr == NULL)
		ioop->inqPtr = dummyInqData;
	if (ioop->dataPtr == NULL)
		ioop->dataPtr = dummyScsiData;

	l = sprintf(foo, "SCSI Error Report =-=-= (%s)\n"
	  "  SCSI_Status=%02Xh (%s)\n"
	  "  Original_CDB[]:",
			ioop->DevIDStr,
			ioop->SCSIStatus,
			ScsiStatusString[ioop->SCSIStatus] );
	l += dump_cdb(foo+l, ioop->cdbPtr);
	if (ScsiOpcodesPtr) {
		const char *opstr = ScsiOpcodesPtr[ioop->cdbPtr[0]];
		l += sprintf(foo+l, " - \"%s\"", opstr);
	}
	l += sprintf(foo+l, "\n  SenseData[%02Xh]:", 8+SD_Additional_Sense_Length(ioop->sensePtr));
	l += dump_sd(foo+l, ioop->sensePtr);
	l += sprintf(foo+l, "\n  SenseKey=%Xh (%s); FRU=%02Xh\n  ASC/ASCQ=%02Xh/%02Xh",
			sk, skstr, SD_FRU(ioop->sensePtr), asc, ascq );

	if ((ascqStr = ascq_lookup(asc, ascq)) != NULL) {
		if (ascqStr[0] != '(')
			l += sprintf(foo+l, " \"%s\"", ascqStr);
		else
			l += sprintf(foo+l, " %s", ascqStr);
	}

//	if (SPECIAL_ASCQ(asc,ascq))
//		l += sprintf(foo+l, " (%02Xh)", ascq);

	PrintF(("%s\n", foo));

	return l;
}

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	init_module - MPI SCSI driver initialization routine
 *
 *	Returns 0 for success, non-zero for failure.
 */
int
#ifdef MODULE
init_module(void)
#else
my_init(void)
#endif
{
	show_mpimod_ver(my_NAME, my_VERSION);

	ScsiDoneCtx = mpi_register(ScsiDone);
	ScsiTaskCtx = mpi_register(ScsiTaskMgmtComplete);

	Q_INIT(&mpiscsi_taskQ, MPI_FRAME_HDR);
	spin_lock_init(&mpiscsi_taskQ_lock);

#ifdef MODULE
	mpi_scsi_host_template.module = &__this_module;
	scsi_register_module(MODULE_SCSI_HA, &mpi_scsi_host_template);
	if (! mpi_scsi_host_template.present) {
		scsi_unregister_module(MODULE_SCSI_HA, &mpi_scsi_host_template);
		return -EBUSY;
	}
#endif

	return 0;
}

#ifdef MODULE
/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	cleanup_module - Perform SCSI driver module unload cleanup
 *
 *	This routine frees all resources associated with each SCSI host
 *	adapter.
 */
void
cleanup_module(void)
{
	if (mpi_scsi_host_template.present)
		scsi_unregister_module(MODULE_SCSI_HA, &mpi_scsi_host_template);

	mpi_deregister(ScsiDoneCtx);
	mpi_deregister(ScsiTaskCtx);
}

#else

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/**
 *	mpiscsi_init - MPI SCSI driver initialization routine when driver is
 *			compiled into the kernel (not as a module)
 */
void __init
mpiscsi_init(void)
{
	my_init();
}

#endif

/*=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

